/*****************************************************************************
 * common.h: misc common functions
 *****************************************************************************
 * Copyright (C) 2003-2014 x264 project
 *
 * Authors: Laurent Aimar <fenrir@via.ecp.fr>
 *          Loren Merritt <lorenm@u.washington.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 * This program is also available under a commercial proprietary license.
 * For more information, contact us at licensing@x264.com.
 *****************************************************************************/

#ifndef X264_COMMON_H
#define X264_COMMON_H


#include "osdep.h"
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <xmmintrin.h>


union  _x264_union128_sse_t { __m128 i; uint64_t a[2]; uint32_t b[4]; uint16_t c[8]; uint8_t d[16]; };
typedef union  _x264_union128_sse_t x264_union128_sse_t;

uint32_t pack16to32_mask( int a, int b );


enum macroblock_position_e
{
    MB_LEFT = 0x01,
    MB_TOP = 0x02,
    MB_TOPRIGHT = 0x04,
    MB_TOPLEFT = 0x08,

    MB_PRIVATE = 0x10,

    ALL_NEIGHBORS = 0xf,
};

enum mb_class_e
{
    I_4x4 = 0,
    I_8x8 = 1,
    I_16x16 = 2,
    I_PCM = 3,

    P_L0 = 4,
    P_8x8 = 5,
    P_SKIP = 6,

    B_DIRECT = 7,
    B_L0_L0 = 8,
    B_L0_L1 = 9,
    B_L0_BI = 10,
    B_L1_L0 = 11,
    B_L1_L1 = 12,
    B_L1_BI = 13,
    B_BI_L0 = 14,
    B_BI_L1 = 15,
    B_BI_BI = 16,
    B_8x8 = 17,
    B_SKIP = 18,

    X264_MBTYPE_MAX = 19
};

enum mb_partition_e
{

    D_L0_4x4 = 0,
    D_L0_8x4 = 1,
    D_L0_4x8 = 2,
    D_L0_8x8 = 3,


    D_L1_4x4 = 4,
    D_L1_8x4 = 5,
    D_L1_4x8 = 6,
    D_L1_8x8 = 7,

    D_BI_4x4 = 8,
    D_BI_8x4 = 9,
    D_BI_4x8 = 10,
    D_BI_8x8 = 11,
    D_DIRECT_8x8 = 12,


    D_8x8 = 13,
    D_16x8 = 14,
    D_8x16 = 15,
    D_16x16 = 16,
    X264_PARTTYPE_MAX = 17,
};


enum cabac_ctx_block_cat_e
{
    DCT_LUMA_DC = 0,
    DCT_LUMA_AC = 1,
    DCT_LUMA_4x4 = 2,
    DCT_CHROMA_DC = 3,
    DCT_CHROMA_AC = 4,
    DCT_LUMA_8x8 = 5,
    DCT_CHROMAU_DC = 6,
    DCT_CHROMAU_AC = 7,
    DCT_CHROMAU_4x4 = 8,
    DCT_CHROMAU_8x8 = 9,
    DCT_CHROMAV_DC = 10,
    DCT_CHROMAV_AC = 11,
    DCT_CHROMAV_4x4 = 12,
    DCT_CHROMAV_8x8 = 13,
};


/****************************************************************************
 * Macros
 ****************************************************************************/
#define X264_MIN(a,b) ( (a)<(b) ? (a) : (b) )
#define X264_MAX(a,b) ( (a)>(b) ? (a) : (b) )
#define X264_MIN3(a,b,c) X264_MIN((a),X264_MIN((b),(c)))
#define X264_MAX3(a,b,c) X264_MAX((a),X264_MAX((b),(c)))
#define X264_MIN4(a,b,c,d) X264_MIN((a),X264_MIN3((b),(c),(d)))
#define X264_MAX4(a,b,c,d) X264_MAX((a),X264_MAX3((b),(c),(d)))
#define XCHG(type,a,b) do{ type t = a; a = b; b = t; } while(0)
#define IS_DISPOSABLE(type) ( type == X264_TYPE_B )
#define FIX8(f) ((int)(f*(1<<8)+.5))
#define ALIGN(x,a) (((x)+((a)-1))&~((a)-1))
#define ARRAY_ELEMS(a) ((sizeof(a))/(sizeof(a[0])))

#define CHECKED_MALLOC( var, size )\
do {\
    var = x264_malloc( size );\
    if( !var )\
        goto fail;\
} while( 0 )
#define CHECKED_MALLOCZERO( var, size )\
do {\
    CHECKED_MALLOC( var, size );\
    memset( var, 0, size );\
} while( 0 )

/* Macros for merging multiple allocations into a single large malloc, for improved
 * use with huge pages. */

/* Needs to be enough to contain any set of buffers that use combined allocations */
#define PREALLOC_BUF_SIZE 1024

#define PREALLOC_INIT\
    int    prealloc_idx = 0;\
    size_t prealloc_size = 0;\
    uint8_t **preallocs[PREALLOC_BUF_SIZE];

#define PREALLOC( var, size )\
do {\
    var = (void*)prealloc_size;\
    preallocs[prealloc_idx++] = (uint8_t**)&var;\
    prealloc_size += ALIGN(size, NATIVE_ALIGN);\
} while(0)

#define PREALLOC_END( ptr )\
do {\
    CHECKED_MALLOC( ptr, prealloc_size );\
    while( prealloc_idx-- )\
        *preallocs[prealloc_idx] += (intptr_t)ptr;\
} while(0)

#define ARRAY_SIZE(array)  (sizeof(array)/sizeof(array[0]))

#define X264_BFRAME_MAX 16
#define X264_REF_MAX 16
#define X264_THREAD_MAX 128
#define X264_LOOKAHEAD_THREAD_MAX 16
#define X264_PCM_COST (FRAME_SIZE(256*BIT_DEPTH)+16)
#define X264_LOOKAHEAD_MAX 250
#define QP_BD_OFFSET (6*(BIT_DEPTH-8))
#define QP_MAX_SPEC (51+QP_BD_OFFSET)
#define QP_MAX (QP_MAX_SPEC+18)
#define QP_MAX_MAX (51+2*6+18)
#define PIXEL_MAX ((1 << BIT_DEPTH)-1)
// arbitrary, but low because SATD scores are 1/4 normal
#define X264_LOOKAHEAD_QP (12+QP_BD_OFFSET)
#define SPEC_QP(x) X264_MIN((x), QP_MAX_SPEC)

// number of pixels (per thread) in progress at any given time.
// 16 for the macroblock in progress + 3 for deblocking + 3 for motion compensation filter + 2 for extra safety
#define X264_THREAD_HEIGHT 24

/* WEIGHTP_FAKE is set when mb_tree & psy are enabled, but normal weightp is disabled
 * (such as in baseline). It checks for fades in lookahead and adjusts qp accordingly
 * to increase quality. Defined as (-1) so that if(i_weighted_pred > 0) is true only when
 * real weights are being used. */

#define X264_WEIGHTP_FAKE (-1)

#define NALU_OVERHEAD 5 // startcode + NAL type costs 5 bytes per frame
#define FILLER_OVERHEAD (NALU_OVERHEAD+1)
#define SEI_OVERHEAD (NALU_OVERHEAD - (h->param.b_annexb && !h->param.i_avcintra_class && (h->out.i_nal-1)))


#define MB_INTERLACED 0
#define SLICE_MBAFF 0
#define PARAM_INTERLACED 0


#define CHROMA_FORMAT h->sps->i_chroma_format_idc
#define CHROMA_H_SHIFT h->mb.chroma_h_shift
#define CHROMA_V_SHIFT h->mb.chroma_v_shift

#define CHROMA_SIZE(s) ((s)>>(CHROMA_H_SHIFT+CHROMA_V_SHIFT))
#define FRAME_SIZE(s) ((s)+2*CHROMA_SIZE(s))
#define CHROMA444 (CHROMA_FORMAT == CHROMA_444)

/* Unions for type-punning.
 * Mn: load or store n bits, aligned, native-endian
 * CPn: copy n bits, aligned, native-endian
 * we don't use memcpy for CPn because memcpy's args aren't assumed to be aligned */

#define M16(src) (((x264_union16_t*)(src))->i)
#define M32(src) (((x264_union32_t*)(src))->i)
#define M64(src) (((x264_union64_t*)(src))->i)
#define M128(src) (((x264_union128_t*)(src))->i)
#define M128_ZERO ((x264_uint128_t){{0,0}})
#define CP16(dst,src) M16(dst) = M16(src)
#define CP32(dst,src) M32(dst) = M32(src)
#define CP64(dst,src) M64(dst) = M64(src)
#define CP128(dst,src) M128(dst) = M128(src)


#define PIXEL_SPLAT_X4(x) ((x)*0x01010101U)
#define MPIXEL_X4(src) M32(src)


#define BIT_DEPTH X264_BIT_DEPTH

#define CPPIXEL_X4(dst,src) MPIXEL_X4(dst) = MPIXEL_X4(src)

#define X264_SCAN8_LUMA_SIZE (5*8)
#define X264_SCAN8_SIZE (X264_SCAN8_LUMA_SIZE*3)
#define X264_SCAN8_0 (4+1*8)

/* Scan8 organization:
 *    0 1 2 3 4 5 6 7
 * 0  DY    y y y y y
 * 1        y Y Y Y Y
 * 2        y Y Y Y Y
 * 3        y Y Y Y Y
 * 4        y Y Y Y Y
 * 5  DU    u u u u u
 * 6        u U U U U
 * 7        u U U U U
 * 8        u U U U U
 * 9        u U U U U
 * 10 DV    v v v v v
 * 11       v V V V V
 * 12       v V V V V
 * 13       v V V V V
 * 14       v V V V V
 * DY/DU/DV are for luma/chroma DC.
 */

#define LUMA_DC   48
#define CHROMA_DC 49

#define FENC_STRIDE 16
#define FDEC_STRIDE 32

union _x264_union16_t{
    uint16_t i;
    uint8_t c[2];
};

typedef union _x264_union16_t x264_union16_t;

union _x264_union32_t{
    uint32_t i;
    uint16_t b[2];
    uint8_t c[4];
};

typedef union _x264_union32_t x264_union32_t;

union _x264_union64_t{
    uint64_t i;
    uint32_t a[2];
    uint16_t b[4];
    uint8_t c[8];
} ;

typedef union _x264_union64_t x264_union64_t;

struct _x264_uint128_t{
    uint64_t i[2];
} ;

typedef struct _x264_uint128_t x264_uint128_t;

union _x264_union128_t {
    x264_uint128_t i;
    uint64_t a[2];
    uint32_t b[4];
    uint16_t c[8];
    uint8_t d[16];
};

typedef union _x264_union128_t x264_union128_t;


typedef uint8_t pixel;
typedef uint32_t pixel4;
typedef int16_t dctcoef;
typedef uint16_t udctcoef;




extern const uint8_t x264_scan8[16 * 3 + 3];

#include "x264.h"
#include "opencl.h"
#include "cabac.h"
#include "bitstream.h"
#include "set.h"
#include "predict.h"
#include "pixel.h"
#include "mc.h"
#include "frame.h"
#include "dct.h"
#include "quant.h"
#include "cpu.h"
#include "threadpool.h"

/****************************************************************************
 * General functions
 ****************************************************************************/
/* x264_malloc : will do or emulate a memalign
 * you have to use x264_free for buffers allocated with x264_malloc */
void *x264_malloc(int);
void x264_free(void *);

/* x264_slurp_file: malloc space for the whole file and read it */
char *x264_slurp_file(const char *filename);

/* mdate: return the current date in microsecond */
int64_t x264_mdate(void);

/* x264_param2string: return a (malloced) string containing most of
 * the encoding options */
char *x264_param2string(x264_param_t *p, int b_res);

/* log */
void x264_log(x264_t *h, int i_level, const char *psz_fmt, ...);

void x264_reduce_fraction(uint32_t *n, uint32_t *d);
void x264_reduce_fraction64(uint64_t *n, uint64_t *d);
void x264_cavlc_init(x264_t *h);
void x264_cabac_init(x264_t *h);

pixel x264_clip_pixel(int x);

int x264_clip3(int v, int i_min, int i_max);

double x264_clip3f(double v, double f_min, double f_max);

int x264_median(int a, int b, int c);

void x264_median_mv(int16_t *dst, int16_t *a, int16_t *b, int16_t *c);

int x264_predictor_difference(int16_t(*mvc)[2], intptr_t i_mvc);

uint16_t x264_cabac_mvd_sum(uint8_t *mvdleft, uint8_t *mvdtop);




extern const uint8_t x264_exp2_lut[64];
extern const float x264_log2_lut[128];
extern const float x264_log2_lz_lut[32];



int x264_exp2fix8(float x);

float x264_log2(uint32_t x);




enum slice_type_e {
    SLICE_TYPE_P = 0,
    SLICE_TYPE_B = 1,
    SLICE_TYPE_I = 2,
};

extern const char slice_type_to_char[3];

enum sei_payload_type_e {
    SEI_BUFFERING_PERIOD = 0,
    SEI_PIC_TIMING = 1,
    SEI_PAN_SCAN_RECT = 2,
    SEI_FILLER = 3,
    SEI_USER_DATA_REGISTERED = 4,
    SEI_USER_DATA_UNREGISTERED = 5,
    SEI_RECOVERY_POINT = 6,
    SEI_DEC_REF_PIC_MARKING = 7,
    SEI_FRAME_PACKING = 45,
};

struct _x264_slice_header_t{
    x264_sps_t *sps;
    x264_pps_t *pps;

    int i_type;
    int i_first_mb;
    int i_last_mb;

    int i_pps_id;

    int i_frame_num;

    int b_mbaff;
    int b_field_pic;
    int b_bottom_field;

    int i_idr_pic_id;

    int i_poc;
    int i_delta_poc_bottom;

    int i_delta_poc[2];
    int i_redundant_pic_cnt;

    int b_direct_spatial_mv_pred;

    int b_num_ref_idx_override;
    int i_num_ref_idx_l0_active;
    int i_num_ref_idx_l1_active;

    int b_ref_pic_list_reordering[2];

    struct {
        int idc;
        int arg;
    } ref_pic_list_order[2][16];


    int b_weighted_pred;
    x264_weight_t weight[16 * 2][3];

    int i_mmco_remove_from_end;
    int i_mmco_command_count;

    struct {
        int i_difference_of_pic_nums;
        int i_poc;
    } mmco[16];

    int i_cabac_init_idc;

    int i_qp;
    int i_qp_delta;
    int b_sp_for_swidth;
    int i_qs_delta;


    int i_disable_deblocking_filter_idc;
    int i_alpha_c0_offset;
    int i_beta_offset;

};

typedef struct _x264_slice_header_t x264_slice_header_t;

struct x264_lookahead_t {
    volatile uint8_t b_exit_thread;
    uint8_t b_thread_active;
    uint8_t b_analyse_keyframe;
    int i_last_keyframe;
    int i_slicetype_length;
    x264_frame_t *last_nonb;
    int thread_handle;
    x264_sync_frame_list_t ifbuf;
    x264_sync_frame_list_t next;
    x264_sync_frame_list_t ofbuf;
};

typedef struct x264_lookahead_t x264_lookahead_t;

typedef struct x264_ratecontrol_t x264_ratecontrol_t;

struct x264_left_table_t {
    uint8_t intra[4];
    uint8_t nnz[4];
    uint8_t nnz_chroma[4];
    uint8_t mv[4];
    uint8_t ref[4];
};

typedef struct x264_left_table_t x264_left_table_t;

struct _x264_frame_stat_t{

    int i_mv_bits;

    int i_tex_bits;

    int i_misc_bits;

    int i_mb_count[19];
    int i_mb_count_i;
    int i_mb_count_p;
    int i_mb_count_skip;
    int i_mb_count_8x8dct[2];
    int i_mb_count_ref[2][16 * 2];
    int i_mb_partition[17];
    int i_mb_cbp[6];
    int i_mb_pred_mode[4][13];
    int i_mb_field[3];

    int i_direct_score[2];

    int64_t i_ssd[3];
    double f_ssim;
    int i_ssim_cnt;
};


typedef struct _x264_frame_stat_t x264_frame_stat_t;


struct x264_t {

    x264_param_t param;

    x264_t *thread[128 + 1];
    x264_t *lookahead_thread[16];
    int b_thread_active;
    int i_thread_phase;
    int i_thread_idx;
    int i_threadslice_start;
    int i_threadslice_end;
    int i_threadslice_pass;
    x264_threadpool_t *threadpool;
    x264_threadpool_t *lookaheadpool;
    int mutex;
    int cv;


    struct {
        int i_nal;
        int i_nals_allocated;
        x264_nal_t *nal;
        int i_bitstream;
        uint8_t *p_bitstream;
        bs_t bs;
    } out;

    uint8_t *nal_buffer;
    int nal_buffer_size;

    x264_t *reconfig_h;
    int reconfig;




    int i_frame;
    int i_frame_num;

    int i_thread_frames;

    int i_nal_type;
    int i_nal_ref_idc;

    int64_t i_disp_fields;
    int i_disp_fields_last_frame;
    int64_t i_prev_duration;
    int64_t i_coded_fields;
    int64_t i_cpb_delay;

    int64_t i_coded_fields_lookahead;
    int64_t i_cpb_delay_lookahead;

    int64_t i_cpb_delay_pir_offset;
    int64_t i_cpb_delay_pir_offset_next;

    int b_queued_intra_refresh;
    int64_t i_last_idr_pts;

    int i_idr_pic_id;


    int (*dequant4_mf[4])[16];
    int (*dequant8_mf[4])[64];

    int (*unquant4_mf[4])[16];
    int (*unquant8_mf[4])[64];

    udctcoef(*quant4_mf[4])[16];
    udctcoef(*quant8_mf[4])[64];
    udctcoef(*quant4_bias[4])[16];
    udctcoef(*quant8_bias[4])[64];
    udctcoef(*quant4_bias0[4])[16];
    udctcoef(*quant8_bias0[4])[64];
    udctcoef(*nr_offset_emergency)[4][64];


    uint16_t *cost_mv[((51+(6*(8 -8)))+18) + 1];
    uint16_t *cost_mv_fpel[((51+(6*(8 -8)))+18) + 1][4];

    const uint8_t *chroma_qp_table;


    x264_slice_header_t sh;


    x264_sps_t sps[1];
    x264_pps_t pps[1];


    int b_sh_backup;
    x264_slice_header_t sh_backup;


    x264_cabac_t cabac;

    struct {

        x264_frame_t **current;

        x264_frame_t **unused[2];


        x264_frame_t **blank_unused;


        x264_frame_t *reference[16 + 2];

        int i_last_keyframe;
        int i_last_idr;
        int i_poc_last_open_gop;



        int i_input;

        int i_max_dpb;
        int i_max_ref0;
        int i_max_ref1;
        int i_delay;
        int i_bframe_delay;
        int64_t i_bframe_delay_time;
        int64_t i_first_pts;
        int64_t i_prev_reordered_pts[2];
        int64_t i_largest_pts;
        int64_t i_second_largest_pts;
        int b_have_lowres;
        int b_have_sub8x8_esa;
    } frames;


    x264_frame_t *fenc;


    x264_frame_t *fdec;


    int i_ref[2];
    x264_frame_t *fref[2][16 + 3];
    x264_frame_t *fref_nearest[2];
    int b_ref_reorder[2];


    int initial_cpb_removal_delay;
    int initial_cpb_removal_delay_offset;
    int64_t i_reordered_pts_delay;


    struct {
        dctcoef luma16x16_dc[3][16];
        dctcoef chroma_dc[2][8];

        dctcoef luma8x8[12][64];
        dctcoef luma4x4[16 * 3][16];
    } dct;


    struct {
        int i_mb_width;
        int i_mb_height;
        int i_mb_count;


        int chroma_h_shift;
        int chroma_v_shift;


        int i_mb_stride;
        int i_b8_stride;
        int i_b4_stride;
        int left_b8[2];
        int left_b4[2];


        int i_mb_x;
        int i_mb_y;
        int i_mb_xy;
        int i_b8_xy;
        int i_b4_xy;


        int i_me_method;
        int i_subpel_refine;
        int b_chroma_me;
        int b_trellis;
        int b_noise_reduction;
        int b_dct_decimate;
        int i_psy_rd;
        int i_psy_trellis;

        int b_interlaced;
        int b_adaptive_mbaff;


        int mv_min[2];
        int mv_max[2];
        int mv_miny_row[3];
        int mv_maxy_row[3];


        int mv_min_spel[2];
        int mv_max_spel[2];
        int mv_miny_spel_row[3];
        int mv_maxy_spel_row[3];

        int16_t mv_limit_fpel[2][2];
        int mv_miny_fpel_row[3];
        int mv_maxy_fpel_row[3];


        unsigned int i_neighbour;
        unsigned int i_neighbour8[4];
        unsigned int i_neighbour4[16];
        unsigned int i_neighbour_intra;
        unsigned int i_neighbour_frame;
        int i_mb_type_top;
        int i_mb_type_left[2];
        int i_mb_type_topleft;
        int i_mb_type_topright;
        int i_mb_prev_xy;
        int i_mb_left_xy[2];
        int i_mb_top_xy;
        int i_mb_topleft_xy;
        int i_mb_topright_xy;
        int i_mb_top_y;
        int i_mb_topleft_y;
        int i_mb_topright_y;
        const x264_left_table_t *left_index_table;
        int i_mb_top_mbpair_xy;
        int topleft_partition;
        int b_allow_skip;
        int field_decoding_flag;






        uint8_t *base;
        int8_t *type;
        uint8_t *partition;
        int8_t *qp;
        int16_t *cbp;
        int8_t(*intra4x4_pred_mode)[8];

        uint8_t(*non_zero_count)[16 * 3];
        int8_t *chroma_pred_mode;
        int16_t(*mv[2])[2];
        uint8_t(*mvd[2])[8][2];
        int8_t *ref[2];
        int16_t(*mvr[2][16 * 2])[2];
        int8_t *skipbp;
        int8_t *mb_transform_size;
        uint16_t *slice_table;

        uint8_t *field;


        pixel *p_weight_buf[16];


        int i_type;
        int i_partition;
        uint8_t i_sub_partition[4];
        int b_transform_8x8;

        int i_cbp_luma;
        int i_cbp_chroma;

        int i_intra16x16_pred_mode;
        int i_chroma_pred_mode;





        int i_skip_intra;


        int b_skip_mc;

        int b_reencode_mb;
        int ip_offset;
        int b_deblock_rdo;
        int b_overflow;

        struct {



            pixel fenc_buf[48 * 16];
            pixel fdec_buf[52 * 32];


            pixel i4x4_fdec_buf[16 * 16];
            pixel i8x8_fdec_buf[16 * 16];
            dctcoef i8x8_dct_buf[3][64];
            dctcoef i4x4_dct_buf[15][16];
            uint32_t i4x4_nnz_buf[4];
            uint32_t i8x8_nnz_buf[4];
            int i4x4_cbp;
            int i8x8_cbp;


            dctcoef fenc_dct8[4][64];
            dctcoef fenc_dct4[16][16];


            uint64_t fenc_hadamard_cache[9];
            uint32_t fenc_satd_cache[32];


            pixel *p_fenc[3];

            pixel *p_fenc_plane[3];


            pixel *p_fdec[3];


            int i_fref[2];

            pixel *p_fref[2][16 * 2][12];
            pixel *p_fref_w[16 * 2];
            uint16_t *p_integral[2][16];


            int i_stride[3];
        } pic;


        struct {

            int8_t intra4x4_pred_mode[(5*8)];


            uint8_t non_zero_count[((5*8)*3)];


            int8_t ref[2][(5*8)];


            int16_t mv[2][(5*8)][2];
            uint8_t mvd[2][(5*8)][2];


            int8_t skip[(5*8)];

            int16_t direct_mv[2][4][2];
            int8_t direct_ref[2][4];
            int direct_partition;
            int16_t pskip_mv[2];


            int i_neighbour_transform_size;
            int i_neighbour_skip;


            int i_cbp_top;
            int i_cbp_left;


            int16_t topright_mv[2][3][2];
            int8_t topright_ref[2][3];


            uint8_t(*deblock_strength)[8][4];
        } cache;


        int i_qp;
        int i_chroma_qp;
        int i_last_qp;
        int i_last_dqp;
        int b_variable_qp;
        int b_lossless;
        int b_direct_auto_read;
        int b_direct_auto_write;


        int i_trellis_lambda2[2][2];
        int i_psy_rd_lambda;
        int i_chroma_lambda2_offset;


        int16_t dist_scale_factor_buf[2][2][16 * 2][4];
        int16_t(*dist_scale_factor)[4];
        int8_t bipred_weight_buf[2][2][16 * 2][4];
        int8_t(*bipred_weight)[4];


        int8_t map_col_to_list0[16 + 2];
        int ref_blind_dupe;
        int8_t deblock_ref_table[16 * 2 + 2];

    } mb;


    x264_ratecontrol_t *rc;


    struct {

        x264_frame_stat_t frame;




        int i_frame_count[3];
        int64_t i_frame_size[3];
        double f_frame_qp[3];
        int i_consecutive_bframes[16 + 1];

        double f_ssd_global[3];
        double f_psnr_average[3];
        double f_psnr_mean_y[3];
        double f_psnr_mean_u[3];
        double f_psnr_mean_v[3];
        double f_ssim_mean_y[3];
        double f_frame_duration[3];

        int64_t i_mb_count[3][19];
        int64_t i_mb_partition[2][17];
        int64_t i_mb_count_8x8dct[2];
        int64_t i_mb_count_ref[2][2][16 * 2];
        int64_t i_mb_cbp[6];
        int64_t i_mb_pred_mode[4][13];
        int64_t i_mb_field[3];

        int i_direct_score[2];
        int i_direct_frames[2];

        int i_wpred[2];

    } stat;


    udctcoef(*nr_offset)[64];
    uint32_t(*nr_residual_sum)[64];
    uint32_t *nr_count;

    udctcoef nr_offset_denoise[4][64];
    uint32_t nr_residual_sum_buf[2][4][64];
    uint32_t nr_count_buf[2][4];

    uint8_t luma2chroma_pixel[7];


    void *scratch_buffer;
    void *scratch_buffer2;
    pixel *intra_border_backup[5][3];


    uint8_t(*deblock_strength[2])[2][8][4];


    x264_predict_t predict_16x16[4 + 3];
    x264_predict8x8_t predict_8x8[9 + 3];
    x264_predict_t predict_4x4[9 + 3];
    x264_predict_t predict_chroma[4 + 3];
    x264_predict_t predict_8x8c[4 + 3];
    x264_predict_t predict_8x16c[4 + 3];
    x264_predict_8x8_filter_t predict_8x8_filter;

    x264_pixel_function_t pixf;
    x264_mc_functions_t mc;
    x264_dct_function_t dctf;
    x264_zigzag_function_t zigzagf;
    x264_zigzag_function_t zigzagf_interlaced;
    x264_zigzag_function_t zigzagf_progressive;
    x264_quant_function_t quantf;
    x264_deblock_function_t loopf;
    x264_bitstream_function_t bsf;

    x264_lookahead_t *lookahead;
    x264_opencl_t opencl;
};





int x264_predictor_roundclip(int16_t(*dst)[2], int16_t(*mvc)[2], int i_mvc, int16_t mv_limit[2][2], uint32_t pmv);

int x264_predictor_clip(int16_t(*dst)[2], int16_t(*mvc)[2], int i_mvc, int16_t mv_limit[2][2], uint32_t pmv);


#endif

