
#include "input.h"

#define FAIL_IF_ERROR( cond, ... ) FAIL_IF_ERR( cond, "raw", __VA_ARGS__ )
struct _raw_hnd_t {
    FILE *fh;
    int next_frame;
    uint64_t plane_size [4];
    uint64_t frame_size;
    int bit_depth;
};
typedef struct _raw_hnd_t raw_hnd_t;

int open_file_RAW (char *psz_filename, hnd_t *p_handle, video_info_t *info, cli_input_opt_t *opt) {
    raw_hnd_t *h = calloc (1, sizeof (raw_hnd_t));
    if (!h)
        return -1;
    if (!opt->resolution) {
        for (char *p = psz_filename; *p; p++)
            if (*p >= '0' && *p <= '9' && sscanf (p, "%dx%d", &info->width, &info->height) == 2)
                break;
    }
    else
        sscanf (opt->resolution, "%dx%d", &info->width, &info->height);
    FAIL_IF_ERROR (! info -> width || ! info -> height, "raw input requires a resolution.\n")
    if (opt->colorspace) {
        for (info->csp = X264_CSP_CLI_MAX - 1; info->csp > X264_CSP_NONE; info->csp--) {
            if (x264_cli_csps[info->csp].name && !strcasecmp (x264_cli_csps[info->csp].name, opt->colorspace))
                break;
        }
        FAIL_IF_ERROR (info -> csp == X264_CSP_NONE, "unsupported colorspace `%s'\n", opt -> colorspace);
    }
    else
        info->csp = X264_CSP_I420;
    h->bit_depth = opt->bit_depth;
    FAIL_IF_ERROR (h -> bit_depth < 8 || h -> bit_depth > 16, "unsupported bit depth `%d'\n", h -> bit_depth);
    if (h->bit_depth > 8)
        info->csp |= X264_CSP_HIGH_DEPTH;
    if (!strcmp (psz_filename, "-"))
        h->fh = stdin;
    else
        h->fh = x264_fopen (psz_filename, "rb");
    if (h->fh == NULL)
        return -1;
    info->thread_safe = 1;
    info->num_frames = 0;
    info->vfr = 0;
    const x264_cli_csp_t *csp = x264_cli_get_csp (info->csp);
    for (int i = 0; i < csp->planes; i++) {
        h->plane_size[i] = x264_cli_pic_plane_size (info->csp, info->width, info->height, i);
        h->frame_size += h->plane_size[i];
        h->plane_size[i] /= x264_cli_csp_depth_factor (info->csp);
    }
    if (x264_is_regular_file (h->fh)) {
        fseek (h -> fh, 0, SEEK_END);
        uint64_t size = ftell (h->fh);
        fseek (h -> fh, 0, SEEK_SET);
        info->num_frames = size / h->frame_size;
    }
    *p_handle = h;
    return 0;
}

int read_frame_internal_RAW (cli_pic_t *pic, raw_hnd_t *h, int bit_depth_uc) {
    int error = 0;
    int pixel_depth = x264_cli_csp_depth_factor (pic->img.csp);
    for (int i = 0; i < pic->img.planes && !error; i++) {
        error |= fread (pic->img.plane[i], pixel_depth, h->plane_size[i], h->fh) != h->plane_size[i];
        if (bit_depth_uc) {
            uint16_t *plane = (uint16_t *) pic->img.plane[i];
            uint64_t pixel_count = h->plane_size[i];
            int lshift = 16 - h->bit_depth;
            for (uint64_t j = 0; j < pixel_count; j++)
                plane[j] = plane[j] << lshift;
        }
    }
    return error;
}

int read_frame_RAW (cli_pic_t *pic, hnd_t handle, int i_frame) {
    raw_hnd_t *h = handle;
    if (i_frame > h->next_frame) {
        if (x264_is_regular_file (h->fh))
            fseek (h->fh, i_frame *h->frame_size, SEEK_SET);
        else
            while (i_frame > h->next_frame) {
                if (read_frame_internal_RAW (pic, h, 0))
                    return -1;
                h->next_frame++;
            }
    }
    if (read_frame_internal_RAW (pic, h, h->bit_depth & 7))
        return -1;
    h->next_frame = i_frame + 1;
    return 0;
}

int close_file_RAW (hnd_t handle) {
    raw_hnd_t *h = handle;
    if (!h || !h->fh)
        return 0;
    fclose (h -> fh);
    free (h);
    return 0;
}

const cli_input_t raw_input = {open_file_RAW, x264_cli_pic_alloc, read_frame_RAW, NULL, x264_cli_pic_clean, close_file_RAW};
